#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <inttypes.h>
#include "../integrate.h"

struct dw_helper {
  float sm1, sm2, dsm_min, dsm_max, sm_scatter_const;
};

#define ERF_CACHE_MIN -8
#define ERF_CACHE_MAX 8
#define ERF_CACHE_BPUNIT 128
#define ERF_CACHE_NUM (((ERF_CACHE_MAX-ERF_CACHE_MIN)*ERF_CACHE_BPUNIT)+2)
float erf_cache[ERF_CACHE_NUM];
#define SIG_CACHE_MIN 0.01
#define SIG_CACHE_MAX 0.99
#define SIG_CACHE_BPUNIT 100
#define SIG_CACHE_NUM 99
float sig_cache[SIG_CACHE_NUM];

double perc_to_normal(double p) {
  double x1 = -14;
  double x2 = 14;
  while (x2-x1 > 1e-7) {
    double half = 0.5*(x1+x2);
    double perc = 0.5*(1.0+erf(half));
    if (perc > p) { x2 = half; }
    else { x1 = half; }    
  }
  return ((x1+x2)*M_SQRT1_2);
}


void gen_erfcache(void) {
  int64_t i;
  for (i=0; i<ERF_CACHE_NUM; i++) {
    double x = (double)ERF_CACHE_MIN + (double)i/(double)(ERF_CACHE_BPUNIT);
    erf_cache[i] = 0.5+0.5*erf(x/M_SQRT2);
  }

  for (i=0; i<SIG_CACHE_NUM; i++) {
    double perc = SIG_CACHE_MIN + (double)i / (double)SIG_CACHE_BPUNIT;
    sig_cache[i] = perc_to_normal(perc);
  }
}

float cached_rank(float x) {
  if (x<ERF_CACHE_MIN) return 0;
  if (x>=ERF_CACHE_MAX) return 1;
  float f = (x-(float)ERF_CACHE_MIN)*(float)ERF_CACHE_BPUNIT;
  int64_t b = f;
  f -= b;
  return (erf_cache[b]+f*(erf_cache[b+1]-erf_cache[b]));
}

float fraction_above_thresh(float x, float thresh, float scatter) {
  return cached_rank((x-thresh)/scatter);
}

float fraction_below_thresh(float x, float thresh, float scatter) {
  return cached_rank((thresh-x)/scatter);
}


double density_weight_integral(double x, void *extra_data) {
  struct dw_helper *d = extra_data;
  float dsm = (x-d->sm1)*d->sm_scatter_const;
  float psm = expf(-0.5*dsm*dsm);
  float psm2 = cached_rank((d->sm2-(x+d->dsm_min))*d->sm_scatter_const)-
    cached_rank((d->sm2-(x+d->dsm_max))*d->sm_scatter_const);
  return psm*psm2;
}

float integrate_density_weight(float sm1, float sm2, float sm_min, float sm_max, float dsm_min, float dsm_max, float sm_scatter_const, float sm_weight) {
  //Integrates P(SM2>(SM1+dsm_min)&SM2<(SM1+dsm_max)|SM1)P(SM1=x) for x=sm_min to x=sm_max
  struct dw_helper d={0};
  d.sm1 = sm1;
  d.sm2 = sm2;
  d.dsm_min = dsm_min;
  d.dsm_max = dsm_max;
  d.sm_scatter_const = sm_scatter_const;
  /*  float est_sm_w = (sm1 < sm_min) ? density_weight_integral(sm_min) :
      density_weight_integral(sm_max);*/
  float tol = sm_weight*1e-5;
  if (!(tol>0)) tol=1e-5;
  float val=adaptiveSimpsons(density_weight_integral, &d,
			     sm_min, sm_max, tol, 10);
  float ret = val*sm_scatter_const*(0.5*M_2_SQRTPI*M_SQRT1_2)/sm_weight;

  //1/sqrt(2*pi*scatter^2)*integral
  if (!isfinite(ret)) ret=0;
  if (ret>1) ret=1;
  return ret; 
}


double approx(float sm1, float sm2) {
  //double params[5]={1.024948e+01, 2.357396e-01, 1.843233e-01, 2.498690e-01, 7.187006e-02}; //sig=0.07
  double params[5]={1.024807e+01, 2.222332e-01, 2.043431e-01, 2.502337e-01, 1.027049e-01};   //sig=0.1
  double mid = params[0];
  double width = params[1];
  double trans = params[2];
  double sep = params[3];
  double width_min = params[4];
  double dsm = (sm1-mid)/width;
  double width2 = width_min*(1.0+(sqrt(2.0)-1.0)*expf(-dsm*dsm));

  double midp = (mid-width)+2.0*width*cached_rank((sm1-mid)/trans);
  return cached_rank((sm2-(midp-sep))/width2)-
    cached_rank((sm2-(midp+sep))/width2);
}

int main(void) {
  float sm_scatter_const = 1.0/0.1;
  gen_erfcache();
  double m1=9.2,m2=9.2;
  //struct point *p = NULL;
  //int64_t num_p = 
  for (m1=9.5; m1<10.7; m1+=0.01) {
    for (m2=9.2; m2<11.2; m2+=0.01) {
      float sm_weight = cached_rank((m1-10.0)*sm_scatter_const)-
	cached_rank((m1-10.5)*sm_scatter_const);
      //v = integrate_density_weight(m1, m2, 10, 10.5, -0.25, 0.25, sm_scatter_const, sm_weight);
      //check_realloc_every*p
      double v = integrate_density_weight(m1, m2, 10, 10.5, -0.25, 0.25, sm_scatter_const, sm_weight);
      double v2 = approx(m1, m2);
      double diff = fabs(v-v2);
      printf("%f %f %e %e %e\n", m1, m2, v, v2, diff);
      //printf("%f %f %e %e %e\n", m1, m2, v, 1e-5, 1e-5); //v2, diff);
    }
  }
  return 0;
}
